#include <stdio.h>
#include <string.h>
#include <domainkeys.h>
#include <unistd.h>

int optf = 0;

errorout(DK *dk,DK_STAT st) {
  if (optf && dk) fprintf(stderr, "%s(%d):",dk_errfile(dk),dk_errline(dk));
  fprintf(stderr, "dktest: %s\n", DK_STAT_to_string(st));
  exit(1);
}

main(int argc, char *argv[]) {
  char inbuf[1024];
  size_t inlen;
  char advice[2048];
  size_t advicelen = sizeof(advice);
  DK *dk;
  DK_LIB *dklib;
  char *dkptr = strrchr(argv[0], 'k');
  DK_STAT st;
  char ch;
  int opts=0, optv=0, optt=0, optc=DK_CANON_SIMPLE;
  char *canon = "simple";
  char *keyfn;
  char *selector;
  char privkey[2048];
  FILE *privkeyf;
  size_t privkeylen;
  DK_FLAGS dkf;

  while (1) {
    ch = getopt(argc, argv,"s:vt:fb:c:");
    if (ch == -1) break;
    switch (ch) {
    case 'v':
      optv = 1;
      break;
    case 'f':
      optf = 1;
      break;
    case 's':
      opts = 1;
      keyfn = optarg;
      selector = optarg;
      while (*optarg) {
	if (*optarg == '/')
	  selector = optarg+1;
	optarg++;
      }
      break;
    case 't':
      optt = atoi(optarg);
      break;
    case 'b':
      advicelen = atoi(optarg);
      if (advicelen > sizeof(advice)) advicelen = sizeof(advice);
      break;
    case 'c':
      if (!strcmp(optarg, "simple")) optc = DK_CANON_SIMPLE, canon = "simple";
      else if (!strcmp(optarg, "nofws")) optc = DK_CANON_NOFWS, canon = "nofws";
      else {
	fprintf(stderr, "dktest: unrecognized canonicalization.\n");
	exit(1);
      }
    }
  }

  if (opts) {
    privkeyf = fopen(keyfn, "r");
    if (!privkeyf) { /*TC10*/
      fprintf(stderr, "dktest: can't open private key file %s\n", keyfn);
      exit(1);
    }
    privkeylen = fread(privkey, 1, sizeof(privkey), privkeyf);
    if (privkeylen == sizeof(privkey)) { /* TC9 */
      fprintf(stderr, "dktest: private key buffer isn't big enough, use a smaller private key or recompile.\n");
      exit(1);
    }
    privkey[privkeylen] = '\0';
    fclose(privkeyf);
  }

  if (optt == 1) errorout(NULL,0); /*TC2*/
  if (optt == 2) errorout(NULL,32767); /*TC3*/

  dklib = dk_init(&st);
  if (st != DK_STAT_OK) errorout(NULL,st);
  if (!dklib) errorout(NULL, 200);

  if (optv) {
    dk = dk_verify(dklib, &st);
    if (st != DK_STAT_OK) errorout(dk,st);
  } else if (opts) {
    dk = dk_sign(dklib, &st, optc);
    if (st != DK_STAT_OK) errorout(dk,st);
  } else {
    fprintf(stderr, "dktest: [-f] [-b#] [-v|-s selector] [-t#]\n"); /* TC1 */
    exit(1);
  }

  if (optt == 3) errorout(dk,dk_message(NULL, "", 1)); /* TC4 */
  if (optt == 4) errorout(dk,dk_message(dk, NULL, 1)); /* TC5 */
  if (optt == 5) errorout(dk,dk_message(dk, "", 0)); /* TC6 */

  st = DK_STAT_OK;
  while (1) {
    char *inp;

    inlen = fread(inbuf, 1, sizeof(inbuf), stdin);
    inp = inbuf;
    while (inlen--) {
      if (*inp == '\n') {
	if (st==DK_STAT_OK) st = dk_message(dk, "\r\n", 2);
      } else {
	if (st==DK_STAT_OK) st = dk_message(dk, inp, 1);
      }
      inp++;
    }
    if (inp-inbuf < sizeof(inbuf)) break;
  }
  if (st==DK_STAT_OK) st = dk_eom(dk, &dkf);
  if (optt == 6 && optv) {
    printf("flags: ");
    if (dkf & DK_FLAG_TESTING) printf("t");
    if (dkf & DK_FLAG_SIGNSALL) printf("s");
    printf("\n");
  } else if (optt == 6 && opts) {
    errorout(dk, dk_getsig(dk, NULL, NULL, advicelen)); /* TC14 */
  } else if (optt == 7) {
    printf("%s\n",dk_from(dk));	/* TC14-1 */
  } else if (optt == 8 && opts) {
    dk_getsig(dk, privkey, advice, advicelen);
    if (st != DK_STAT_OK) errorout(dk,st);
    printf("%d %d\n",dk_siglen(privkey), strlen(advice)); /* TC39 */
  } else if (opts) {
    if (st != DK_STAT_OK) errorout(dk,st);
    st = dk_getsig(dk, privkey, advice, advicelen);
    if (st != DK_STAT_OK) errorout(dk,st);
    printf("Comment: DomainKeys? See http://antispam.yahoo.com/domainkeys\n"
	   "DomainKey-Signature: a=rsa-sha1; q=dns; c=%s;\n"
	   "  s=%s; d=%s;\n"
	   "  b=%s;\n", canon, selector, dk_from(dk), advice);
  } else if (optv) {
    char *status;

    switch(st) {
    case DK_STAT_OK: status = "good"; break;
    case DK_STAT_BADSIG: status = "bad"; break;
    case DK_STAT_NOSIG: status = "no signature"; break;
    case DK_STAT_NOKEY:
    case DK_STAT_CANTVRFY: status = "no key"; break;
    case DK_STAT_BADKEY: status = "bad key"; break;
    case DK_STAT_INTERNAL:
    case DK_STAT_ARGS:
    case DK_STAT_SYNTAX: status = "bad format"; break;
    case DK_STAT_NORESOURCE: status = "no resources"; break;
    case DK_STAT_REVOKED: status = "revoked"; break;
    }
    printf("Comment: DomainKeys? See http://antispam.yahoo.com/domainkeys\n"
	   "DomainKey-Status: %s\n", status);
    rewind(stdin);
  }
  if (st != DK_STAT_OK) errorout(dk,st);
}
