/* This was originally http://netsounds.eresmas.net/spfquery/spfquery.c,
 * but it isn't anymore. */

#include <stdio.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>

static unsigned short getshort(unsigned char *cp) {
  return (cp[0] << 8) | cp[1];
}

/* we return a null-terminated string which is either e={temp,perm} or else
 * a malloc'ed copy of the txt record.  That means that you have to call this
 * routine in a stereotypical fashion:

	  txtrec = dns_txt(query);
	  if (!strcmp(txtrec,"e=perm;")) {
	    STUFF
          } else if (!strcmp(txtrec,"e=temp;")) {
	    STUFF
	  } else {
	    STUFF
	    free(txtrec);
          }
 * Note that it never returns NULL, and note that you only free it if it's
 * not one of the errors.
 */
char *dns_txt(char *dn)
{
    u_char response[PACKETSZ+1]; /* response */
    int responselen;		/* buffer length */

    int i, rc;			/* misc variables */
    int ancount, qdcount;	/* answer count and query count */
    u_short type, rdlength;		/* fields of records returned */
    u_char *eom, *cp;

    u_char buf[PACKETSZ+1];	/* we're storing a TXT record here, not just a DNAME */
    u_char *bufptr;

    responselen = res_query(dn, C_IN, T_TXT, response, sizeof(response));
    if (responselen  < 0){
      if (h_errno == TRY_AGAIN) return "e=temp;";
      else return "e=perm;";
    }

    qdcount = getshort( response + 4); /* http://crynwr.com/rfc1035/rfc1035.html#4.1.1. */
    ancount = getshort( response + 6);

    eom = response + responselen;
    cp  = response + HFIXEDSZ;

    while( qdcount-- > 0 && cp < eom ) {
      rc = dn_expand( response, eom, cp, (char *)buf, MAXDNAME );
      if( rc < 0 ) {
	return "e=perm;";
      }
      cp += rc + QFIXEDSZ;
    }

    while( ancount-- > 0 && cp < eom ) {
      rc = dn_expand( response, eom, cp, (char *)buf, MAXDNAME );
      if( rc < 0 ) {
	return "e=perm;";
      }
      
      cp += rc;

      if (cp + RRFIXEDSZ >= eom) return "e=perm;";

      type = getshort(cp + 0); /* http://crynwr.com/rfc1035/rfc1035.html#4.1.3. */
      rdlength = getshort(cp + 8);
      cp += RRFIXEDSZ;
      
      if( type != T_TXT ) {
	cp += rdlength;
	continue;
      }
      
      bufptr = buf;
      while (rdlength && cp < eom) {
	int cnt;

	cnt = *cp++;		 /* http://crynwr.com/rfc1035/rfc1035.html#3.3.14. */
	if( bufptr-buf + cnt + 1 >= PACKETSZ )
	  return "e=perm;";
	if (cp + cnt > eom)
	  return "e=perm;";
	memcpy( bufptr, cp, cnt);
	rdlength -= cnt + 1;
	bufptr += cnt;
	cp += cnt;
	*bufptr = '\0';
      }
      
      return (char *) strdup( buf );
    }
    return "e=perm;";
}
